
#include "trithorpe.h"


Expr::~Expr() { }

err Expr::beginPut() { return Str("bad assignment"); }
err Expr::finishPut() { return Str("bad assignment"); }
err Expr::emitCall() { return Str("bad function call"); }
bool Expr::isError() { return false; }
err Expr::getError() { return NIL; }
err Expr::emitAsABlock() { return piler->outputAsABlock(this); }


ConsExpr::ConsExpr(Compiler* c, Expr* _head, ConsExpr* _tail) : Expr(c), head(_head), tail(_tail) {
    what= tri('C','o','n','s');
    DBG(stderr, "ConsExpr::ConsExpr: \n" ); 
}
ConsExpr::~ConsExpr() { delete head; delete tail; }
err ConsExpr::emitGet() { 
    tri e;
    if (tail) {
	e= tail->emitGet();    // recurse to emit backwards
	if (e != NIL) return e;

#if TODO
	out( tri('p','o','p') );  // remove all but the final (head) result
#endif
    }

    e= head->emitGet();
    return e;
}
void ConsExpr::print() {
    fprintf(stderr,"ConsExpr(");
    if (tail) tail->print();
    fprintf(stderr," :: ");
    head->print();
    fprintf(stderr,") ");
}


Verbatim::Verbatim(Compiler* c, tri _verb) : Expr(c), verb(_verb) {
    what= tri('V','e','r','b');
    DBG(stderr, "Verbatim::Verbatim: tok=%x\n", verb.peek() ); 
}
Verbatim::~Verbatim() { }
err Verbatim::emitGet() { out( verb ); return NIL; }
void Verbatim::print() { fprintf(stderr,"Verbatim("); verb.to(stderr); fprintf(stderr,") "); }



Lit::Lit(Compiler* c, tri _lit) : Expr(c), lit(_lit) {
    what= tri('L','i','t');
    DBG(stderr, "Lit::Lit: tok=%x\n", lit.peek() ); 
}
Lit::~Lit() { }
err Lit::emitGet() {
    if ( ! lit.isInt() ) out( tri('i','m','m','/') );  // int does not need imm/ ; others do
    out( lit );
    return NIL;
}
void Lit::print() { fprintf(stderr,"Lit("); lit.to(stderr); fprintf(stderr,") "); }



Error::Error(Compiler* c, err _err, Expr* _inner) : Lit(c, _err) , inner(_inner) {
    what= tri('E','r','r','o');
    DBG(stderr, "Error::Error: tok=%x\n", lit.peek() ); 
}
Error::~Error() { }
err Error::emitGet() {
    if (lit.isSym()) out( tri('i','m','m','/') );
    out( lit );
    return lit; // return the err!
}
bool Error::isError() { return true; }
err Error::getError() { return lit; }
void Error::print() {
    fprintf(stderr,"Error("); 
    if (inner) inner->print();
    fprintf(stderr," :: ");
    lit.to(stderr); 
    fprintf(stderr,") ");
}



LocalVar::LocalVar(Compiler* c, int _index)
    : Expr(c)
    , index(_index)
{
    what= tri('L','v','a','r');
}

LocalVar::~LocalVar() { }
err LocalVar::beginPut() { /*nothing*/ return NIL; }
err LocalVar::finishPut() { out( tri('s','t','o','/') ); out(index); return NIL; }
err LocalVar::emitGet() { out( tri('r','c','l','/') ); out(index); return NIL; }
void LocalVar::print() { fprintf(stderr,"Var(%d) ", index); }



Slot::Slot(Compiler* c, Expr* _table, tri _key)
    : Expr(c)
    , table(_table)
    , key(_key)
{
    what= tri('S','l','o','t');
}

Slot::~Slot() {
    delete table;
}
err Slot::beginPut() {
    table->emitGet();
    out(tri('i','m','m','/')); out(key);
    return NIL;
}
err Slot::finishPut() { out( tri('a','p','u','t') ); return NIL; }
err Slot::emitGet() { 
    table->emitGet();
    out(tri('i','m','m','/')); out(key);
    out( tri('a','g','e','t') );
    return NIL;
}
err Slot::emitCall() { 
    table->emitGet();
    out(tri('i','m','m','/')); out(key);
    return NIL;
}
void Slot::print() { 
    fprintf(stderr,"Slot{ ");
    table->print();
    fprintf(stderr," . ");
    key.to(stderr);
    fprintf(stderr," } ");
}



Assign::Assign(Compiler* c, Expr* _lhs, Expr* _rhs)
    : Expr(c)
    , lhs(_lhs)
    , rhs(_rhs)
{
    what= tri('A','s','s','i');
}

Assign::~Assign() {
    delete lhs;
    delete rhs;
}
err Assign::emitGet() { 
    lhs->beginPut();
    rhs->emitGet();
    lhs->finishPut();
    return NIL;
}
void Assign::print() { 
    fprintf(stderr,"Slot{ ");
    lhs->print();
    fprintf(stderr," := ");
    rhs->print();
    fprintf(stderr," } ");
}



Binop::Binop(Compiler* c, tri _op, Expr* _one, Expr* _two)
    : Expr(c)
    , op(_op)
    , one(_one)
    , two(_two)
{
    what= tri('B','i','n','o');
}

Binop::~Binop() { delete one; delete two; }
err Binop::emitGet() { one->emitGet(); two->emitGet(); out(op); return NIL; }
void Binop::print() { 
    fprintf(stderr, "Binop(");
    one->print(),
    fprintf(stderr, " ");
    op.to(stderr),
    fprintf(stderr, " ");
    two->print(),
    fprintf(stderr, ") ");
}

Call::Call(Compiler* c, Expr* _one, Expr* _two)
    : Expr(c)
    , one(_one)
    , two(_two)
{
    what= tri('C','a','l','l');
}

Call::~Call() { delete one; delete two; }
err Call::emitGet() {
//DBG(stderr,"Call::emitGet -> ");
//print();
    one->emitCall();
    two->emitGet();
    out( tri('c','a','l','/') );
    out( tri(3) );
    return NIL;
}
void Call::print() { 
    fprintf(stderr, "Call(");
    one->print(),
    two->print(),
    fprintf(stderr, ") ");
}

IfExpr::IfExpr(Compiler* c, Expr* _cond, Expr* _thenBlock, Expr* _elseBlock)
    : Expr(c)
    , cond(_cond)
    , thenBlock(_thenBlock)
    , elseBlock(_elseBlock)
{
    what= tri('I','f','E','x');
}

IfExpr::~IfExpr() { delete cond; delete thenBlock; delete elseBlock; }

err IfExpr::emitGet() {
DBG(stderr,"IfExpr::emitGet -> ");
if (verbose) print();
    cond->emitGet();
    out( tri('i','f') );

    if (thenBlock)
	thenBlock->emitAsABlock();
    else
	out( NIL );

    if (elseBlock)
	elseBlock->emitAsABlock();
    else
	out( NIL );

    return NIL;
}
void IfExpr::print() { 
    fprintf(stderr, "IfExpr(");
    cond->print();
    if (thenBlock) thenBlock->print(); else say << "(no THEN)";
    if (elseBlock) elseBlock->print(); else say << "(no ELSE)";
    fprintf(stderr, ") ");
}

LoopExpr::LoopExpr(Compiler* c, Expr* _body)
    : Expr(c)
    , body(_body)
{
    what= tri('L','o','o','p');
}

LoopExpr::~LoopExpr() { delete body; }

err LoopExpr::emitGet() {
DBG(stderr,"IfExpr::emitGet -> ");
print();

    out( tri('l','o','o','p') );
    body->emitAsABlock();
    
    return NIL;
}
void LoopExpr::print() { 
    fprintf(stderr, "LoopExpr(");
    body->print(),
    fprintf(stderr, ") ");
}


/////////////////////////////////////////////////////////////////

int Compiler::addLocal( tri varName ) 
{
    int already= findLocal( varName );
    if ( already >= 0 ) return already;

    localVars.aAdd( varName ); 
    return localVars.aSize()-1;
}
int Compiler::findLocal( tri varName )
{
    if ( ! useLocal() ) return -1;

    int n= localVars.aSize();
    for (int i=0; i<n; i++) {
	if ( localVars.aGet(i) == varName ) return i;
    }
    return -1;
}


Compiler::Compiler( tri _src, tri _dest, tri _lobby )
    : src( _src )
    , isize( src.aSize() )
    , inext( 0 )
    , dest( _dest )
    , lobby( _lobby )
{
    //localVars.setNewTab(); // no locals at outer
}


void Compiler::output(const tri& x) {
    //sync();
    //fprintf(stderr, "COMPILER::OUTPUT %x tab %x  ADD %x\n", dest.peek(), dest.asTab(), x.peek() );
    CheckDeep( x );
    dest.aAdd(x);
    CheckDeep( dest );
}
void Expr::out(const tri& x) { piler->output(x); }

Expr* Compiler::parseExpr( )
{
    Expr* y= 0;
    Expr* x= parseBasic();

    if (!x) return 0;

    while ( ! done() ) {
	tri tok= cur();

	switch (tok.peek()) {


	case SYM1('.'):
	{
	    advance();
	    if ( done() ) goto bail;

	    tri key= cur();
	    advance();

	    x= new Slot(this, x, key);
	} break;

	case SYM1('('):
	{
	    advance();
	    if ( done() ) goto bail;

	    //while ( cur() != ')' ) {
		y= parseExpr();
		if (!y) goto bail;
	    //}

	    MUST ( cur() == tri(')') );
	    advance();

	    x= new Call(this, x, y);
	} break;

	case SYM1('='):
	{
	    advance();
	    if ( done() ) goto bail;

	    y= parseBasic();
	    if (!y) goto bail;

	    x= new Assign(this, x, y);
	} break;


	case SYM1('+'):
	case SYM1('-'):
	case SYM1('*'):
	case SYM1('/'):
	case SYM1('%'):
	case SYM1('&'):
	case SYM1('|'):
	case SYM1('^'):
	case SYM2('=','='):
	case SYM2('!','='):
	case SYM2('<','='):
	case SYM2('>','='):
	case SYM1('<'):
	case SYM1('>'):
	{
	    advance();
	    if ( done() ) goto bail;

	    y= parseBasic();
	    if (!y) goto bail;

	    x= new Binop(this, tok, x, y);
	} break;

	default:
	    goto finish;
	}
    }
finish:
    return x;
bail:
    if (x) delete x;
    if (y) delete y;
    return 0;
}

Expr* Compiler::parseBasic()
{
    while (1) {
	if ( done() )  return new Error(this, "unexpected END in parseBasic") ; // EOS

	tri tok= cur();
	advance();

DBG(stderr, "PT: tok=%x\n", tok.peek() ); 

	if ( tok.isInt() ) {
	    return new Lit(this, tok);
	} else if ( tok.isTab() ) {
	    return new Lit(this, tok);

	} else if ( tok == tri('f','u','n') ) {

	    return parseFun();

	} else if ( tok == tri('i','f') ) {

	    return parseIf();

	} else if ( tok == tri('l','o','o','p') ) {

	    return parseLoop();

	} else if ( tok == tri('b','r','k') ) {
	    return new Verbatim( this, tri('b','r','k') );
	} else if ( tok == tri('b','r','e','a') ) {
	    return new Verbatim( this, tri('b','r','k') );

	} else if ( tok == tri('n','i','l') ) {
	    
	    return new Verbatim( this, tri('n','i','l') );

	} else if ( tok == tri('H','o','m','e') ) {
	    
	    return new Verbatim( this, tri('h','o','m','e') );

	} else if ( tok == tri('(') ) {
	    Expr* z= parseExpr();
	    EQ( tri(')'), cur() );
	    advance();
	    return z;

	} else if ( tok.isSym() ) {

	    // if it is followed by '=', make it a local
	    if ( useLocal() && cur() == tri('=') )  addLocal( tok );
		
	    int i= findLocal( tok );
	    if (0<=i) {
		return new LocalVar( this, i );
	    } else {
		Expr* LOBBY= new Verbatim( this, tri('h','o','m','e') );
		return new Slot( this, LOBBY, tok );
	    }
	}

	return 0;
    }
}

Expr* Compiler::parseExprList()
{
    ConsExpr* z= NULL;

	while ( ! done()  &&  cur() != tri(')')  &&  cur() != tri('}') ) {

	    if ( cur() == tri(';') || cur() == tri(',') ) {
		advance();
		continue;
	    }

	    Expr* expr= parseExpr();
	    if (!expr) {
		if (z) delete z;
		return new Error(this, Str("bad Expr in Expr List") );
	    }
	    if (expr->isError()) {
		if (z) delete z;
		return new Error(this, Str("... in Expr List"), expr );
	    }

	    if (verbose) expr->print();

	    z= new ConsExpr(this, expr, z);
	}

    return z;  // may be NULL if empty block
}


Expr* Compiler::parseBlock()
{

	if ( cur() != tri('{') )   return new Error(this, Str("`{' expected before Block") );
	advance();

	Expr* z= parseExprList();

	if (z->isError()) return new Error(this, Str("... in Block"), z );

	if ( cur() != tri('}') )   {
	    delete z;
	    return new Error(this, Str("`}' expected after Block") );
	}
	advance();

	return z;  // may be NULL if empty block
}

Expr* Compiler::parseIf()
{
    Expr* c= parseExpr();
    if (c->isError()) { return new Error(this, "... in condition of if", c); }

    Expr* t= parseBlock();
    if (t->isError()) { delete c; return new Error(this, "... in then-branch of if", t); }

    Expr* e= NULL;

    if ( cur()== tri('e','l','s','e') ) {
	advance();

	e= parseBlock();
	if (e->isError()) { delete t; delete c; return new Error(this, "... in else-branch of if", e); }
    } else {
	say << "SKIPPING ELSE because cur is " << cur() << "\n";
    }

    return new IfExpr( this, c, t, e );
}

Expr* Compiler::parseLoop()
{
    tri z;  z.setNewTab();

    Expr* body= parseBlock();
    if (body->isError()) { return new Error(this, "... in body of Loop", body); }
    
    return new LoopExpr( this, body );
}

err Compiler::outputAsABlock(Expr* expr)
{
    err e;
    {
    tri z("tcod");
CheckDeep(z); CheckDeep(dest);
    {
	tri saved_dest= dest;
	dest= z;
CheckDeep(z); CheckDeep(dest);
	e= expr->emitGet();
CheckDeep(z); CheckDeep(dest);
	dest= saved_dest;
CheckDeep(z); CheckDeep(dest);
    }

CheckDeep(z); CheckDeep(dest);
    if ( e == NIL ) { output( z ); } else { say << "ERROR IN outputAsABlock:" << e << "\n"; }
    }
  CheckDeep(dest);

    return e;
}

Expr* Compiler::parseFun()
{
    tri z;  z.setNewTab();

	if ( cur() != tri('(') )   return new Error(this, Str("`fun' not followed by `('") );
	advance();

	tri saveLocalVars= localVars;
	localVars.setNewTab();
	addLocal( tri('t','h','i','s') );
	addLocal( tri('t','h','i','m') );

	while ( cur() != tri(')') ) {

	    addLocal( cur() );
	    advance();
	}
	advance();

	z.setNewTab();
	z.aAdd( tri('t','c','o','d') );
	z.aAdd( tri('l','c','l','/') );   // local var info follows
	z.aAdd( tri() );  // to be fixed

	Expr* body= parseBlock();


	if (body) {
	    if (body->isError()) 
		return  new Error(this, Str("... in function definition"), body );

	    // Substitute our z for dest
	    tri saved_dest= dest;
	    dest= z;

	    body->emitGet();

	    dest= saved_dest;
	    delete body;

	} else {

	    z.aAdd( tri('n','i','l') );

	}

	z.aPut( 2, localVars );  // fix with num local args

	localVars= saveLocalVars;  // restore
    return new Lit(this, z);
}
    
tri CompileString(tri x, char* locals )
{
    tri z;
    z.setNewTab();
    Compiler* piler= new Compiler(x,z);
    piler->localVars.setNewTab();

    tri lcls(locals);
    for (int i=0; i<lcls.aSize(); i++) piler->addLocal( lcls.aGet(i) );

    Expr* ex= piler->parseExpr();

    if (!ex) return tri('b','a','d');
    if (! piler->done() ) return tri('m','o','r','e');

    //ex->print();
    ex->emitGet();
    delete ex;
    delete piler;
    return z;
}

