
#include <stdio.h>
extern void SayChar(int c);  // someday all debug thu this

////////////////////  integers

typedef unsigned char u1;
typedef unsigned short u2;
typedef unsigned int u4;
typedef   signed char s1;
typedef   signed short s2;
typedef   signed int s4;

typedef u4 word;  // must be unsigned

////////////////////  asserts

extern "C" {

#include <assert.h>

    void sync(void);  // for breakpoints
    void abort(void);
    char* getenv( const char* );
    int atoi( const char* );

}

extern int expect_error;
extern int count_error;
extern int count_expected_error;


extern int count_new;
extern int count_delete;
extern int count_new_array;
extern int count_delete_array;

extern int verbose;

#define DBG if(verbose)fprintf

#ifdef DEBUG

#define MUST(a) assert_must(a,__STRING(a),__LINE__,__FILE__,__PRETTY_FUNCTION__)
#define CANT(a) assert_must((!(a)),__STRING((!(a))),__LINE__,__FILE__,__PRETTY_FUNCTION__)
#define EQ(a,b) assert_eq(a,b,__STRING(a),__STRING(b),__LINE__,__FILE__,__PRETTY_FUNCTION__)
#define NE(a,b) assert_ne(a,b,__STRING(a),__STRING(b),__LINE__,__FILE__,__PRETTY_FUNCTION__)
#define ISBAD(a) { int old= count_expected_error; expect_error=1; (a); expect_error=0; MUST(old<count_expected_error); }

const int TOOBIG = 100000;  // nothing should get bigger

#else

#define MUST(a) /*nothing*/
#define CANT(a) /*nothing*/
#define EQ(a,b) /*nothing*/
#define NE(a,b) /*nothing*/
#define ISBAD(a) /*nothing*/

const int TOOBIG = 0x0fffFFFF;  // nothing should get bigger

#endif

class tri;
class tab;
class vec;
class TriTerp;
struct Meta;

typedef tri err;  // when NIL means good, and else means error

bool assert_must(bool a, const char* A, int line, const char* file, const char* func);
bool assert_eq(word a, word b, const char* A, const char* B, int line, const char* file, const char* func);
bool assert_ne(word a, word b, const char* A, const char* B, int line, const char* file, const char* func);
bool assert_eq(tri a, tri b, const char* A, const char* B, int line, const char* file, const char* func);
bool assert_ne(tri a, tri b, const char* A, const char* B, int line, const char* file, const char* func);

void Say(const char *s);  // debugging prints
void bomb(const char* problem);   // aborts
void bad(void);                   // breakpoint here
void memcheck(void); 
void CheckDeep(tri x, int depth= 8);

struct Sayer {
	Sayer& operator << (int a);
	Sayer& operator << (char a);
	Sayer& operator << (const char* a);
	Sayer& operator << (const tri& a);
	Sayer& operator << (const tab& a);
	Sayer& operator << (const vec& a);
	Sayer& operator << (const TriTerp& a);
	Sayer& operator << (class Expr* a);
};

extern Sayer say;
extern bool runUnitTests;

///////////////////  classes

#define SYM1(a) ( tri::SYM | (u1)(a) )
#define SYM2(a,b) ( tri::SYM | ((u1)(a)<<8) | (u1)(b) )
#define SYM3(a,b,c) ( tri::SYM | ((u1)(a)<<16) | ((u1)(b)<<8) | (u1)(c) )
#define SYM4(a,b,c,d) ( tri::SYM | ((u1)(a)<<24) | ((u1)(b)<<16) | ((u1)(c)<<8) | (u1)(d) )

#define	NIL_GUTS	SYM3('n','i','l') 
#define	NIL		tri('n','i','l')


/////////////////////////////////////////////////////////

class vec {
private:
    int	    siz;
    int	    cap;
    tri*    d;

    static tri const_zero;

    void do_grow(int goal);
    void grow(int goal) {
	// if we already are at goal, ok
	if (cap<goal) do_grow(goal);
    }
    void operator= (const vec& FORBIDDEN);

public:
    ~vec();
    vec() : siz(0), cap(0) , d(0) {}
    tri& operator [] (int i) const ;
    tri& operator [] (int i) ;
    int capacity() const { return cap; }
    int size() const { return siz; }
};

class tab {

friend class tri;

private:
    int	    refs;

    tab(tri* setMe);
    void operator= (const tab& FORBIDDEN);

public:
    const static word MAGIC= 0xae9b5366;
    word    magic;

    Meta*    meta;

    vec	    arr;
    vec	    dict;

    static long    countCtors;
    static long    countDtors;

    ~tab();
    void link( void* by );
    void unlink( void* by );
    void checkMagic() const;

    int	aSize() const { checkMagic(); return arr.size(); }
    int	dSize() const { checkMagic(); return dict.size(); }

    tri aGet(int i) const;
    void aPut(int i, tri x);

    tri recall(const tri& tindex) const;
    tri recall(word windex) const;

    void store(const tri& tindex, const tri& x) ;
    void store(word windex, const tri& x) ;

    void aAdd(tri item);
};




class tri {  // the tri-modal trithorpe thingy
    

private:
	word  guts;

	static const int NUM_TAG_BITS = 2;

	static const word BASIS = 1 << ( 8 * sizeof(word) - NUM_TAG_BITS );
	static const word LO = BASIS-1;
	static const word HI = ~LO;

public:
	tri() : guts( NIL_GUTS ) { link(); }
	tri(const tri& copyMe) : guts(copyMe.guts) { link(); }
	tri(tri& copyMe) : guts(copyMe.guts) { link(); }

	tri(int x) : guts(LO&x | INT&HI) { }
	tri(word x) : guts(LO&x | INT&HI) { }

	tri(char a) : guts( SYM1(a) ) { MUST(isSym()); }
	tri(char a, char b) : guts( SYM2(a,b) ) { MUST(isSym()); }
	tri(char a, char b, char c) : guts( SYM3(a,b,c) ) { MUST(isSym()); }
	tri(char a, char b, char c, char d) : guts( SYM4(a,b,c,d) ) { MUST(isSym()); }

	tri(const char* tcodeScript);

	static tri NewCodVec(const char*);
	static tri NewDeepCodVec(const char*);

	static const word INT =  BASIS * 0;
	static const word SYM =  BASIS * 1;
	static const word UNK =  BASIS * 2;
	static const word TAB =  BASIS * 3;


	static word isWhat(word w) { return (w&HI); }
	static bool isInt(word w) { return (w&HI) == INT; }
	static bool isSym(word w) { return (w&HI) == SYM; }
	static bool isUnk(word w) { return (w&HI) == UNK; }
	static bool isTab(word w) { return (w&HI) == TAB; }

	word isWhat() const { return (guts&HI); }
	bool isInt() const { return (guts&HI) == INT; }
	bool isSym() const { return (guts&HI) == SYM; }
	bool isUnk() const { return (guts&HI) == UNK; }
	bool isTab() const { return (guts&HI) == TAB; }

	word peek() const { return guts; }

	bool operator == (const tri& that) const { return this->guts == that.guts; }
	bool operator != (const tri& that) const { return this->guts != that.guts; }

	tri& operator = (const tri& copyMe) { unlink(); this->guts= copyMe.guts; link(); return *this; }

	int	aSize() const { return asTab()->aSize(); }
	int	dSize() const { return asTab()->dSize(); }

	tri   aGet(int i) const;
	void  aPut(int i, tri x);
	void  aAdd(tri item);

	void  aStore(tri key, tri value) { asTab()->store(key,value); }
	tri   aRecall(tri key) { return asTab()->recall(key); }


	static word asInt(word w) {

	    MUST( isInt(w) );

	    return w & LO; 
	}

	static tab* asTab(word w) {

	    MUST( isTab(w) );

	    tab* z= (tab*)((w & LO)<<2);
	    z->checkMagic();
	    return z;
	}

	word asInt() const {

	    MUST( isInt() );

	    return guts & LO; 
	}

	tab* asTab() const {

	    MUST( isTab() );

	    tab* z= (tab*)((guts & LO)<<2);
	    z->checkMagic();
	    return z;
	}

	tri& setInt(word x) {
	    unlink();
	    guts= ( x&LO | INT&HI ); 
	    return *this;
	}
	tri& setTab(tab* x) {   // called back from tab::tab
	    x->checkMagic();
	    unlink();

	    MUST( x );

	    guts= ( (u4(x)>>2)&LO | TAB&HI ); 
	    link();
	    return *this;
	}
	tri& setNewTab() {
	    new tab( this );   // will call back, to setTab()
	    return *this;
	}

	tri& setSym(char a) { setSym('@',0,0,a); return *this; }
	tri& setSym(char a, char b) { setSym('@',0,a,b); return *this; }
	tri& setSym(char a, char b, char c) { setSym('@',a,b,c); return *this; }
	tri& setSym(char a, char b, char c, char d);
	tri& setSym(word w);

	tri& setTag(tri tag);

	~tri() { unlink(); }
	void link() { 
	    //DBG(stderr, "tri:link %x %x %x\n", this, guts, guts*4); 
	    if (isTab()) { asTab()->link(this); }
	}
	void unlink() { 
	    //DBG(stderr, "tri:unlink %x %x %x\n", this, guts, guts*4); 
	    if (isTab()) { asTab()->unlink(this); }
	    guts= 0;
	}

    tri recall(const tri& tindex) const ;


    tri& setTCodeItemFromString(const char* s, int begin, int end);
    tri& setTCodeArrayFromString(const char* s, int *returnLastUsed= NULL, bool deeply= true);


    void toString( char buf[] ) const;
    void to( FILE* f ) const;

	tri& operator << (int a);
	tri& operator << (word a);
	tri& operator << (char a);
	tri& operator << (unsigned char a);
	tri& operator << (const char* a);
	tri& operator << (const tri& a);
	tri& operator << (const tab& a);
	tri& operator << (const vec& a);
	tri& operator << (const TriTerp& a);
	tri& operator << (class Expr* a);
};


inline    tri& vec::operator [] (int i) const {
	//grow(i+1);
	if (i<cap) {
	    void* z= (char*)d + i * sizeof(word);
//DBG(stderr, "[] %d<%d : %d\n", i, cap, ((tri*)z)->peek() );
	    return * (tri*) z;
	} else {
//DBG(stderr, "[] %d not< %d : zero\n", i, cap );
		return const_zero;
	}
}

inline    tri& vec::operator [] (int i) {
	grow(i+1);
	if (siz<i+1) siz= i+1;

	void* z= (char*)d + i * sizeof(word);
//DBG(stderr, "[]=== %d %d ( %d )===\n", i, cap, ((tri*)z)->peek() );
	return * (tri*) z;
}

inline tri tab::aGet(int i) const { checkMagic(); return arr[i]; }
inline void tab::aPut(int i, tri x) { checkMagic(); arr[i]= x; }
inline void tab::aAdd(tri item) { checkMagic(); aPut( aSize(), item ); }

inline tri tri::aGet(int i) const { return asTab()->aGet(i); }
inline void tri::aPut(int i, tri x) { asTab()->aPut(i,x); }
inline void tri::aAdd(tri item) { asTab()->aAdd( item ); }

/////////////////////////////////////////////////////////
//
//   Strings.........

bool IsPrintableAsciiObj(tri x);
char* StrDupIfPrintableAsciiObj(tri x);

int StrLen(const char* s);
int StrSum(const char* s);
char* StrDup (const char* s);
int StrCmp(const char* a, const char* b);

inline bool StrEq(const char* a, const char* b) { return !StrCmp(a,b);}

tri	Str(const char* s);        // create new String tri
bool 	IsPritableAsciiObj(tri x);
tri	Intern(const char* s);

struct InternStr {
	tri		obj;
	const char*	str;
	InternStr*	next;

	InternStr(tri _obj, const char* _str, InternStr* _next);
};

class InternTable {

#define INTERN_MASK 255
	InternStr*	bucket[INTERN_MASK+1];
public:
	InternStr*	intern(const char* s);
};

extern InternTable It;

/////////////////////////////////////////////////////////

inline bool IsTrue(tri x) { word w= x.peek(); return w!=0 && w!=SYM3('n','i','l'); }
inline bool IsFalse(tri x) { word w= x.peek(); return w==0 || w==SYM3('n','i','l'); }

// met = meta table for tab
struct Meta {
    tri  tag;
    void (*finalizer)(tab*);
};


#define MAX_FRAMES 100
#define MAX_STACK 2000

struct TriFrame {  // a stack frame 
    static const int STOP= 1;
    static const int CATCH= 2;
    static const int FUNC= 4;
    static const int LOOP= 8;

    int	    fflags;
    tri	    kp;
    int    pc;
    tri*    bp;
    tri	    lobby;

    TriFrame();
};


class TriTerp {  // main inTerpreter structure

    struct TriFrame frames[MAX_FRAMES];

    tri*    sp;
    tri	    stack[MAX_STACK];

public:

    struct TriFrame *f;
    tri	    home;

// constants
    const tri TCODE_MAGIC;

// cotr
    TriTerp();
    void init();    // in case of no constructor

// frames
    void advanceFrame();
    void retreatFrame();

    err	testEval( tri script, tri arg1, tri arg2 );

    // stack operations:
    void push(tri);
    tri pop();
    tri peek(int backN);

private:
    tri doLoop();
    void doInt( const tri& );
    err doSym( tri );
    err doTab( tri );
    err doUnk( word );
    err doCall();
    err doReturn();
    err doStop();
    err doIf();
    err doWhile();
    err doBreak();
    err doPutc();
    err doLclInfo();
    
    tri pcImmediate();  // returns immediate entry at program counter & increments pc

public:
    void dumpStack() const;
};


tri CompileString(tri x, char* locals= "" );
void dumpVec(tri x, const char* s, int depth=4);

///////////////////////////////////////

class Compiler;

class Expr {

public:

    tri	    what;  // in leiu of class object

    Compiler* piler;

    Expr(Compiler* c) : piler(c) {}
    virtual ~Expr();
    virtual err beginPut(); // abort()
    virtual err finishPut(); // abort()
    virtual err emitGet() = 0;
    virtual err emitCall(); // abort()
    virtual bool isError();
    virtual err getError();
    virtual void print() = 0;
    void out(const tri& x);
    err emitAsABlock();
};

class ConsExpr : public Expr {
    Expr*	head;
    ConsExpr*	tail;

public:
    ConsExpr(Compiler*, Expr* _head, ConsExpr* _tail);
    ~ConsExpr();
    err emitGet();
    void print();
};

class Verbatim : public Expr {
    tri	verb;

public:
    Verbatim(Compiler*, tri _verb);
    ~Verbatim();
    err emitGet();
    void print();
};


class Lit : public Expr {
protected:
    tri	lit;

public:
    Lit(Compiler*, tri _lit);
    ~Lit();
    err emitGet();
    void print();
};

class Error : public Lit {
    Expr*	inner;

public:
    Error(Compiler*, err _err, Expr* _Inner= NULL);
    ~Error();
    err emitGet();
    virtual bool isError();
    virtual err getError();
    void print();
};

class LocalVar : public Expr {
    int	    index;

public:
    LocalVar(Compiler*, int _index);
    ~LocalVar();
    err beginPut();
    err finishPut();
    err emitGet();
    void print();
};

class Slot : public Expr {
    Expr*	    table;
    tri		    key;

public:
    Slot(Compiler*, Expr* _table, tri _key);
    ~Slot();
    err beginPut();
    err finishPut();
    err emitGet();
    err emitCall();
    void print();
};

class Assign : public Expr {
    Expr*	    lhs;
    Expr*	    rhs;

public:
    Assign(Compiler*, Expr* _lhs, Expr* _rhs);
    ~Assign();
    err emitGet();
    void print();
};

class Binop : public Expr {
    tri	    op;
    Expr*   one;
    Expr*   two;

public:
    Binop(Compiler*, tri _op, Expr* _a1, Expr* _a2);
    ~Binop();
    err emitGet();
    void print();
};

class IfExpr : public Expr {
    Expr*   cond;
    Expr*   thenBlock;
    Expr*   elseBlock;

public:
    IfExpr(Compiler*, Expr* _cond,  Expr* _thenBlock, Expr* _elseBlock );
    ~IfExpr();
    err emitGet();
    void print();
};

class LoopExpr : public Expr {
    Expr*   body;

public:
    LoopExpr(Compiler*, Expr* _body );
    ~LoopExpr();
    err emitGet();
    void print();
};

class Call : public Expr {
    Expr*   one;
    Expr*   two;

public:
    Call(Compiler*, Expr* _a1, Expr* _a2);
    ~Call();
    err emitGet();
    void print();
};

struct ExprPair {
	err	e;  // for the error status
	Expr*	x;  // for the data

	ExprPair( err _e, Expr* _x ) : e(_e) , x(_x) {}
	ExprPair( const ExprPair& copyMe ) : e(copyMe.e) , x(copyMe.x) {}
	ExprPair& operator = ( const ExprPair& copyMe ) { e= copyMe.e; x= copyMe.x; return *this; }
	~ExprPair() {}
};
	
struct Pair {
	err	e;  // for the error status
	tri	x;  // for the data

	Pair( ) : e() , x() {}
	Pair( err _e, tri _x ) : e(_e) , x(_x) {}
	Pair( const Pair& copyMe ) : e(copyMe.e) , x(copyMe.x) {}
	Pair& operator = ( const Pair& copyMe ) { e= copyMe.e; x= copyMe.x; return *this; }
	~Pair() {}
};
	

class Compiler {

    tri	    src;
    int	    isize;
    int	    inext;

public:
    tri	    dest;
    tri	    lobby;

    tri	    numLocalVars;
    tri	    localVars;

    Compiler( tri source, tri dest, tri lobby= NIL );

    Expr* parseExpr( );
    Expr* parseBasic( );
    Expr* parseFun( );
    Expr* parseExprList( );
    Expr* parseBlock( );
    Expr* parseIf( );
    Expr* parseLoop( );

    err outputAsABlock( Expr* );

    void output(const tri& x);
    tri cur() { return src.aGet(inext); }
    void advance() { 
	if (verbose) say << "{*" << cur() << "*}";
	++inext; 
    }
    bool done() { return inext>=isize; }

    bool useLocal( ) { return localVars != NIL; }
    int addLocal( tri varName );
    int findLocal( tri varName );  // -1 means not found

};

// END
